/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    rt2560.h

    Abstract:
    RT2560 ASIC related definition & structures

    Revision History:
    Who         When          What
    --------    ----------    ----------------------------------------------

*/

#ifndef __RT2560_H__
#define __RT2560_H__

#define CSR3        0x000C      // STA MAC address register 0
#define CSR4        0x0010      // STA MAC address register 1
#define CSR7        0x001C      // Interrupt source register
#define CSR16       0x0040      // TSF timer register 0
#define CSR17       0x0044      // TSF timer register 1

// Statistic Register
#define CNT0        0x00A0      // Dot11 FCS error count

// Security coprocessor registers
#define SECCSR0     0x0028      // WEP control register
#define SECCSR1     0x0158      // WEP control register 

// Transmit related CSRs
#define TXCSR0      0x0060      // TX cintrol register

//
// Tx / Rx / Prio / Atim ring descriptor definition
//
#define DESC_OWN_HOST       0
#define DESC_OWN_NIC        1

// =================================================================================
// TX / RX ring descriptor format
// =================================================================================

//
// TX descriptor format, Tx ring, Atim ring & Priority Ring
//
typedef struct  _TXD_STRUC {
    // Word 0
    ULONG       Owner:1;            // Descriptor owner bit
    ULONG       Valid:1;            // Entry valid bit
    ULONG       TxResult:3;         // Filled by MAC ASIC
    ULONG       RetryCount:3;       // Retry result
    ULONG       MoreFrag:1;         // More fragment following this tx ring
    ULONG       ACK:1;
    ULONG       Timestamp:1;
    ULONG       Ofdm:1;
    ULONG       CipherOwn:1;
    ULONG       IFS:2;
    ULONG       RetryMd:1;
    ULONG       DataByteCnt:12;
    ULONG       Rsv1:1;         //RTS:1;
    ULONG       CipherAlg:3;

    // Word 1
    ULONG       BufferAddressPa;

    // Word 2
    ULONG       IvOffset:6;
    ULONG       Aifs:2;
    ULONG       CWmin:4;
    ULONG       CWmax:4;
    ULONG       Rsv2:16;

    // Word 3
    ULONG       PlcpSignal:8;
    ULONG       PlcpService:8;
    ULONG       PlcpLengthLow:8;
    ULONG       PlcpLengthHigh:8;

    // Word 4
    ULONG       Iv;

    // Word 5
    ULONG       Eiv;

    // Word 6-9
    UCHAR       Key[16];

    // Word 10 - 11 Reserved,   not necessary to put into the structure.
    ULONG       RTS:1;
    ULONG       TxRate:7;   // software use only. keep record of the Tx rate, RATE_1,...
    ULONG       Rsv3:24;

}   TXD_STRUC, *PTXD_STRUC;

//
// Rx descriptor format, Rx Ring
//
typedef struct  _RXD_STRUC  {
    // Word 0
    ULONG       Owner:1;
    ULONG       U2M:1;
    ULONG       Mcast:1;
    ULONG       Bcast:1;
    ULONG       MyBss:1;
    ULONG       Crc:1;
    ULONG       Ofdm:1;
    ULONG       PhyErr:1;
    ULONG       CipherOwner:1;
    ULONG       IcvError:1;
    ULONG       IvOffset:6;
    ULONG       DataByteCnt:12;
    ULONG       Rsv1:1;        // Drop:1;           // Drop this frame after NULL cipher operation
    ULONG       CipherAlg:3;

    // Word 1
    ULONG       BufferAddressPa;

    // Word 2 - 3
    UCHAR       BBR0;
    UCHAR       BBR1;   // suppose to read back RSSI
    UCHAR       TA[6];

    // Word 4
    ULONG       Iv;

    // Word 5
    ULONG       Eiv;
    
    // Word 6-9
    UCHAR       Key[16];
    
    // Word 10 - 11 Reserved,   not necessary to put into the structure.
    ULONG       Drop:1;         // Drop this frame after NULL cipher operation
    ULONG       Rsv2:31;

}   RXD_STRUC, *PRXD_STRUC;

//
// CSR3: STA MAC register 0
//
typedef union   _CSR3_STRUC {
    struct  {
        UCHAR       Byte0;      // MAC address byte 0
        UCHAR       Byte1;      // MAC address byte 1
        UCHAR       Byte2;      // MAC address byte 2
        UCHAR       Byte3;      // MAC address byte 3
    }   field;
    ULONG           word;
}   CSR3_STRUC, *PCSR3_STRUC;

//
// CSR4: STA MAC register 1
//
typedef union   _CSR4_STRUC {
    struct  {
        UCHAR       Byte4;      // MAC address byte 4
        UCHAR       Byte5;      // MAC address byte 5
        UCHAR       Rsvd0;
        UCHAR       Rsvd1;
    }   field;
    ULONG           word;
}   CSR4_STRUC, *PCSR4_STRUC;

//
// CSR7: Interrupt source register
// Write one to clear corresponding bit
//
typedef union   _CSR7_STRUC {
    struct  {
        ULONG       TbcnExpire:1;       // Beacon timer expired interrupt
        ULONG       TwakeExpire:1;      // Wakeup timer expired interrupt
        ULONG       TatimwExpire:1;     // Timer of atim window expired interrupt
        ULONG       TxRingTxDone:1;     // Tx ring transmit done interrupt
        ULONG       AtimRingTxDone:1;   // Atim ring transmit done interrupt
        ULONG       PrioRingTxDone:1;   // Priority ring transmit done interrupt
        ULONG       RxDone:1;           // Receive done interrupt
        ULONG       DecryptionDone:1;   // Decryption done interrupt
        ULONG       EncryptionDone:1;   // Encryption done interrupt
        ULONG       UartTxThreshold:1;      // UART1 TX reaches threshold
        ULONG       UartRxThreshold:1;      // UART1 RX reaches threshold
        ULONG       UartIdleThreshold:1;    // UART1 IDLE over threshold
        ULONG       UartTxBufferError:1;    // UART1 TX buffer error
        ULONG       UartRxBufferError:1;    // UART1 RX buffer error
        ULONG       Uart2TxThreshold:1;     // UART2 TX reaches threshold
        ULONG       Uart2RxThreshold:1;     // UART2 RX reaches threshold
        ULONG       Uart2IdleThreshold:1;   // UART2 IDLE over threshold
        ULONG       Uart2TxBufferError:1;   // UART2 TX buffer error
        ULONG       Uart2RxBufferError:1;   // UART2 RX buffer error
        ULONG       Timecsr3Expired:1;      // TIMECSR3 hardware timer expired (for 802.1H quiet period)
        ULONG       Rsvd:12;
    }   field;
    ULONG           word;
}   CSR7_STRUC, *PCSR7_STRUC, INTSRC_STRUC, *PINTSRC_STRUC;


#endif  // __RT2560_H__

