/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    rtmp_def.h

    Abstract:
    Miniport related definition header

    Revision History:
    Who         When          What
    --------    ----------    ----------------------------------------------
    Paul Lin    08-01-2002    created

*/
#ifndef __RTMP_DEF_H__
#define __RTMP_DEF_H__

//
//  Debug information verbosity: lower values indicate higher urgency
//
#define RT_DEBUG_OFF        0
#define RT_DEBUG_ERROR      1
#define RT_DEBUG_WARN       2
#define RT_DEBUG_TRACE      3
#define RT_DEBUG_INFO       4
#define RT_DEBUG_LOUD       5

#define NIC_TAG             ((ULONG)'0652')
#define NIC_DBG_STRING      ("**RT2560**")

//-------------------------------------------------------------------------
// EEPROM definition
//-------------------------------------------------------------------------
#define NUM_EEPROM_BBP_PARMS        19

//
// Entry number for each DMA descriptor ring
//
#define TX_RING_SIZE            32
#define ATIM_RING_SIZE          4
#define PRIO_RING_SIZE          16    // 8
#define RX_RING_SIZE            32
#define MGMT_RING_SIZE          32
#define RING_DESCRIPTOR_SIZE    48
#define MAX_MCAST_LIST_SIZE     32
#define TX_RING                 0xa
#define PRIO_RING               0xc
#define RX_RING                 0xd
#define MAX_TX_PROCESS          4
#define MAX_RX_PROCESS          4
#define MAX_CLIENT              4
#define MAX_FRAME_SIZE          2346                    // Maximum 802.11 frame size

// Length definitions
#define MAC_ADDR_LEN                      6
#define MAX_LEN_OF_SSID                   32
#define MAX_LEN_OF_KEY                    32      // 32 octets == 256 bits, Redefine for WPA
#define MAX_LEN_OF_CHANNELS               42      // 14 channels @2.4G +  12@UNII + 4 @MMAC + 11 @HiperLAN2 + 1 as NULL termination
#define MAX_LEN_OF_SUPPORTED_RATES        12    // 1, 2, 5.5, 11, 6, 9, 12, 18, 24, 36, 48, 54

#define MAX_LEN_OF_MLME_BUFFER            512
#define MAX_LEN_OF_BSS_TABLE              64
#define MAX_VIE_LEN                       128   // New for WPA cipher suite variable IE sizes.

#define MAX_LEN_OF_MLME_QUEUE            10
#define MAX_LEN_OF_MLME_QUEUE_MSG        512

#define MAX_MLME_HANDLER_MEMORY           20    //each them cantains  MAX_LEN_OF_MLME_BUFFER size 

#define CIPHER_TEXT_LEN                   128

#define ASSOC_FUNC_SIZE                  44 // 4-state * 12-event
#define AUTH_FUNC_SIZE                   15 // 3-state * 5-event
#define AUTH_RSP_FUNC_SIZE               6 // 2-state * 3-event
#define SYNC_FUNC_SIZE                   30  // 3-state * 10-event
#define WPA_PSK_FUNC_SIZE               5

#define fRTMP_ADAPTER_INTERRUPT_IN_USE      0x00000002
#define fRTMP_ADAPTER_RESET_IN_PROGRESS     0x00000080
#define fRTMP_ADAPTER_INTERRUPT_ACTIVE      0x00002000
#define fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS  0x00004000
#define fRTMP_ADAPTER_RADIO_OFF             0x00020000

#define MAX_TX_POWER_LEVEL                100   /* mW */
#define MAX_RTS_THRESHOLD                 2347  /* byte count */
#define MAX_FRAG_THRESHOLD                2346  /* byte count */
#define MIN_FRAG_THRESHOLD                256   /* byte count */

// key related definitions
#define SHARE_KEY_NO                      4
#define MAX_LEN_OF_SHARE_KEY              16
#define PAIRWISE_KEY_NO                   4
#define GROUP_KEY_NO                      4

// power status related definitions
#define PWR_ACTIVE                        0
#define PWR_SAVE                          1

#define MLME_CNTL_STATE_MACHINE    5

//
// rtmp_data.c use these definition
//
#define LENGTH_802_11           24
#define LENGTH_802_3            14
#define LENGTH_802_3_TYPE       2
#define LENGTH_802_1_H          8
#define LENGTH_CRC              4
#define MAX_SEQ_NUMBER          0x0fff

#define SUCCESS_WITHOUT_RETRY   0
#define SUCCESS_WITH_RETRY      1
#define FAIL_RETRY_LIMIT        2
#define FAIL_INVALID            3
#define FAIL_OTHER              4

#define RATE_1                  0
#define RATE_2                  1
#define RATE_5_5                2
#define RATE_11                 3
#define RATE_6                  4   // OFDM
#define RATE_9                  5   // OFDM
#define RATE_12                 6   // OFDM
#define RATE_18                 7   // OFDM
#define RATE_24                 8   // OFDM
#define RATE_36                 9   // OFDM
#define RATE_48                 10  // OFDM
#define RATE_54                 11  // OFDM
#define RATE_72                 12
#define RATE_100                13
#define RATE_FIRST_OFDM_RATE    RATE_6

#define IFS_BACKOFF             0
#define IFS_SIFS                1

#define LONG_RETRY              1
#define SHORT_RETRY             0

#define CIPHER_NONE             0
#define CIPHER_WEP64            1
#define CIPHER_WEP128           2
#define CIPHER_TKIP             3
#define CIPHER_AES              4

// 802.1X controlled port definition
#define WPA_802_1X_PORT_SECURED         1
#define WPA_802_1X_PORT_NOT_SECURED     2

#define PAIRWISE_KEY            1
#define GROUP_KEY               2

//IE code
#define IE_SSID                           0
#define IE_SUPP_RATES                     1
#define IE_FH_PARM                        2
#define IE_DS_PARM                        3
#define IE_CF_PARM                        4
#define IE_TIM                            5
#define IE_IBSS_PARM                      6
#define IE_COUNTRY                        7     // 802.11d
#define IE_802_11D_REQUEST                10    // 802.11d
#define IE_CHALLENGE_TEXT                 16
#define IE_POWER_CONSTRAINT               32    // 802.11h d3.3
#define IE_POWER_CAPABILITY               33    // 802.11h d3.3
#define IE_TPC_REQUEST                    34    // 802.11h d3.3
#define IE_TPC_REPORT                     35    // 802.11h d3.3
#define IE_SUPP_CHANNELS                  36    // 802.11h d3.3
#define IE_CHANNEL_SWITCH_ANNOUNCEMENT    37    // 802.11h d3.3
#define IE_MEASUREMENT_REQUEST            38    // 802.11h d3.3
#define IE_MEASUREMENT_REPORT             39    // 802.11h d3.3
#define IE_QUIET                          40    // 802.11h d3.3
#define IE_IBSS_DFS                       41    // 802.11h d3.3
#define IE_ERP                            42    // 802.11g
#define IE_EXT_SUPP_RATES                 50    // 802.11g
#define IE_WPA                            221   // WPA
#define IE_RSN                            48    // 802.11i d3.0

#define CNTL_FUNC_SIZE                    1

// Control state machine
#define CNTL_IDLE                        100

// BSS Type definitions
#define BSS_INDEP                         0  // = Ndis802_11IBSS
#define BSS_INFRA                         1  // = Ndis802_11Infrastructure
#define BSS_ANY                           2  // = Ndis802_11AutoUnknown

#define CW_MIN_IN_BITS          3         // actual CwMin = 2^CW_MIN_IN_BITS - 1 = 7
#define CW_MAX_IN_BITS          8         // actual CwMax = 2^CW_MAX_IN_BITS - 1 = 255
#define RSSI_TO_DBM_OFFSET      110       // for RT2530 RSSI-100 = dBm

// value domain of MacHdr.tyte, which is b3..b2 of the 1st-byte of MAC header
#define BTYPE_MGMT              0   // 00
#define BTYPE_CNTL              1   // 01
#define BTYPE_DATA              2   // 10

// value domain of MacHdr.subtype, which is b7..4 of the 1st-byte of MAC header
// Management frame
#define SUBTYPE_ASSOC_REQ       0
#define SUBTYPE_ASSOC_RSP       1
#define SUBTYPE_REASSOC_REQ     2
#define SUBTYPE_REASSOC_RSP     3
#define SUBTYPE_PROBE_REQ       4
#define SUBTYPE_PROBE_RSP       5
#define SUBTYPE_BEACON          8
#define SUBTYPE_ATIM            9
#define SUBTYPE_DISASSOC        10
#define SUBTYPE_AUTH            11
#define SUBTYPE_DEAUTH          12
#define SUBTYPE_ACTION          13

// Control Frame
#define SUBTYPE_BLOCK_ACK_REQ   8
#define SUBTYPE_BLOCK_ACK       9
#define SUBTYPE_PS_POLL         10
#define SUBTYPE_RTS             11  // 1011
#define SUBTYPE_CTS             12  // 1100
#define SUBTYPE_ACK             13  // 1101
#define SUBTYPE_CFEND           14
#define SUBTYPE_CFEND_CFACK     15

// Data Frame
#define SUBTYPE_DATA                0
#define SUBTYPE_DATA_CFACK          1
#define SUBTYPE_DATA_CFPOLL         2
#define SUBTYPE_DATA_CFACK_CFPOLL   3
#define SUBTYPE_NULL_FUNC           4
#define SUBTYPE_CFACK               5  // 0101
#define SUBTYPE_CFPOLL              6
#define SUBTYPE_CFACK_CFPOLL        7
#define SUBTYPE_QDATA               8
#define SUBTYPE_QDATA_CFACK         9
#define SUBTYPE_QDATA_CFPOLL        10
#define SUBTYPE_QDATA_CFACK_CFPOLL  11
#define SUBTYPE_QOS_NULL            12
#define SUBTYPE_QOS_CFACK           13
#define SUBTYPE_QOS_CFPOLL          14
#define SUBTYPE_QOS_CFACK_CFPOLL    15

#define DRS_TX_QUALITY_WORST_BOUND      4 

#endif  // __RTMP_DEF_H__

