#ifndef RT_TOOL_H
#define RT_TOOL_H

#include <sys/socket.h>
#include <linux/types.h>
#include <linux/wireless.h>

#define  CONFIGURATION_UI_VERSION   "1.2.0"
#define  CONFIGURATION_UI_DATE      "2004-01-30"
#define  NIC_DEVICE_NAME            "RT2500AP"
#define  RALINK                     "Ralink"

//////////////////////////////////////////////////////////////////////////////////// 
#if WIRELESS_EXT <= 11
#ifndef SIOCDEVPRIVATE
#define SIOCDEVPRIVATE                                  0x8BE0
#endif
#define SIOCIWFIRSTPRIV                                 SIOCDEVPRIVATE
#endif

#define RT_PRIV_IOCTL                                   SIOCIWFIRSTPRIV + 0x01

//
// IEEE 802.11 OIDs
//
// Ralink defined OIDs
#define OID_GET_SET_TOGGLE                              0x8000

#define OID_802_11_BSSID                                0x0101
#define OID_802_11_SSID                                 0x0102
#define OID_802_11_NETWORK_TYPES_SUPPORTED              0x0103
#define OID_802_11_NETWORK_TYPE_IN_USE                  0x0104
#define OID_802_11_TX_POWER_LEVEL                       0x0105
#define OID_802_11_RSSI                                 0x0106
#define OID_802_11_RSSI_TRIGGER                         0x0107
#define OID_802_11_INFRASTRUCTURE_MODE                  0x0108
#define OID_802_11_FRAGMENTATION_THRESHOLD              0x0109
#define OID_802_11_RTS_THRESHOLD                        0x010A
#define OID_802_11_NUMBER_OF_ANTENNAS                   0x010B
#define OID_802_11_RX_ANTENNA_SELECTED                  0x010C
#define OID_802_11_TX_ANTENNA_SELECTED                  0x010D
#define OID_802_11_SUPPORTED_RATES                      0x010E
#define OID_802_11_DESIRED_RATES                        0x010F
#define OID_802_11_CONFIGURATION                        0x0110
#define OID_802_11_STATISTICS                           0x0111
#define OID_802_11_ADD_WEP                              0x0112
#define OID_802_11_REMOVE_WEP                           0x0113
#define OID_802_11_DISASSOCIATE                         0x0114
#define OID_802_11_POWER_MODE                           0x0115
#define OID_802_11_BSSID_LIST                           0x0116
#define OID_802_11_AUTHENTICATION_MODE                  0x0117
#define OID_802_11_PRIVACY_FILTER                       0x0118
#define OID_802_11_BSSID_LIST_SCAN                      0x0119
#define OID_802_11_WEP_STATUS                           0x011A
// Renamed to reflect better the extended set of encryption status
#define OID_802_11_ENCRYPTION_STATUS                    OID_802_11_WEP_STATUS
#define OID_802_11_RELOAD_DEFAULTS                      0x011B
// Added to allow key mapping and default keys
#define OID_802_11_ADD_KEY                              0x011C
#define OID_802_11_REMOVE_KEY                           0x011D
#define OID_802_11_ASSOCIATION_INFORMATION              0x011E
#define OID_802_11_TEST                                 0x011F

#define OID_802_3_CURRENT_ADDRESS                       0x0120
#define OID_GEN_RCV_OK                                  0x0121
#define OID_GEN_RCV_NO_BUFFER                           0x0122
#define OID_GEN_MEDIA_CONNECT_STATUS                    0x0123

#define RT_OID_DEVICE_NAME                              0x0200
#define RT_OID_802_11_PREAMBLE                          0x0201
#define RT_OID_802_11_LINK_STATUS                       0x0202
#define RT_OID_802_11_RESET_COUNTERS                    0x0203
#define RT_OID_802_11_AC_CAM                            0x0204
#if DBG
#define RT_OID_802_11_HARDWARE_REGISTER                 0x0205
#endif
#define RT_OID_802_11_RACONFIG                          0x0206
#define RT_OID_802_11_COUNTRY_REGION                    0x0207
#define    RT_OID_802_11_RADIO                          0x0208
#define RT_OID_802_11_RX_AGC_VGC_TUNING                 0x0209
#define RT_OID_802_11_EVENT_TABLE                       0x0210
#define RT_OID_802_11_MAC_TABLE                         0x0211
#define RT_OID_802_11_PHY_MODE                          0x0212
#define RT_OID_802_11_TX_PACKET_BURST                   0x0213
#define RT_OID_802_11_TURBO_MODE                        0x0214
#define RT_OID_802_11_AP_CONFIG                         0x0215
#define RT_OID_802_11_ACL                               0x0216
#define RT_OID_802_11_STA_CONFIG                        0x0217
#define RT_OID_VERSION_INFO                             0x0218

#define RT_OID_802_11_WDS                               0x0219
#define RT_OID_802_11_RADIUS_DATA                       0x0220
#define RT_OID_802_11_WPA_REKEY                         0x0221


//////////////////////////////////////////////////////////////////////////////////// 
#define RT2500AP_SYSTEM_PATH   "/etc/Wireless/RT2500AP/"
#define RT2500AP_SYSTEM_DATA   "RT2500AP.dat"

#define AUTHNTICATOR_NAME      "rt2500apd"
// value domain for pAd->EventTab.Log[].Event
#define EVENT_RESET_ACCESS_POINT    0 // Log = "hh:mm:ss   Restart Access Point"
#define EVENT_ASSOCIATED            1 // Log = "hh:mm:ss   STA 00:01:02:03:04:05 associated"
#define EVENT_DISASSOCIATED         2 // Log = "hh:mm:ss   STA 00:01:02:03:04:05 left this BSS"
#define EVENT_AGED_OUT              3 // Log = "hh:mm:ss   STA 00:01:02:03:04:05 was aged-out and removed from this BSS"
#define EVENT_COUNTER_M             4
#define EVENT_INVALID_PSK           5
#define EVENT_MAX_EVENT_TYPE        6

typedef long            LONG;
typedef unsigned char   UCHAR;
typedef unsigned short  USHORT;
typedef unsigned long   ULONG;
typedef unsigned int    UINT;

typedef unsigned char*  PUCHAR;
typedef unsigned short* PUSHORT;
typedef int             BOOLEAN;

typedef void            VOID;
typedef __u64           ULONGLONG;
typedef    ULONGLONG    LARGE_INTEGER;

//
//  Debug information verbosity: lower values indicate higher urgency
//
#define RT_DEBUG_OFF        0
#define RT_DEBUG_ERROR      1
#define RT_DEBUG_WARN       2
#define RT_DEBUG_TRACE      3
#define RT_DEBUG_INFO       4
#define RT_DEBUG_LOUD       5

//
//  MACRO for debugging information
//
extern ULONG    RTDebugLevel;

#define DBGPRINT(Level, Fmt)        \
{                                   \
    if (Level <= RTDebugLevel)      \
    {                               \
        qDebug Fmt;                 \
    }                               \
}


typedef struct _REGSTR_COUNTRY_ENTRY
{
    int     nCountry;    // define REGSTR_COUNTRYREGION 
    int     nChDef;      // default channel to use
    int     nChMin;      // minimum channel allowed
    int     nChMax;      // maximum channel allowed
} REGSTR_COUNTRY_ENTRY, *PREGSTR_COUNTRY_ENTRY;

#define REGSTR_COUNTRYREGION_MINIMUM        0   //minimum
#define REGSTR_COUNTRYREGION_FCC            0   // FCC, CH1-11
#define REGSTR_COUNTRYREGION_IC             1   // IC (Canada), CH1-11
#define REGSTR_COUNTRYREGION_ETSI           2   // ETSI, CH1-13
#define REGSTR_COUNTRYREGION_SPAIN          3   // SPAIN, CH10-11
#define REGSTR_COUNTRYREGION_FRANCE         4   // FRANCE, CH10-13
#define REGSTR_COUNTRYREGION_MKK            5   // MKK, CH14
#define REGSTR_COUNTRYREGION_MKK1           6   // MKK1(TELEC), CH1-14
#define REGSTR_COUNTRYREGION_ISRAEL         7   // ISRAEL, CH3-9
#define REGSTR_COUNTRYREGION_MAXIMUM        7   // maximum

typedef struct _PAIR_CHANNEL_FREQ_ENTRY
{
    ULONG    lChannel;
    ULONG    lFreq;
} PAIR_CHANNEL_FREQ_ENTRY, *PPAIR_CHANNEL_FREQ_ENTRY;

typedef UCHAR   NDIS_802_11_MAC_ADDRESS[6];

typedef struct _RT_PROFILE_SETTING {
    UCHAR       Profile[32+1];
    UCHAR       SSID[32+1];
    UINT        SsidLen;
    UINT        Channel;
    UINT        Authentication;
    UINT        Encryption;
    UINT        NetworkType;
    UINT        KeyIndex;
    UINT        KeyLength;
    UINT        KeyType;
    UCHAR       Key1[26+1];
    UCHAR       Key2[26+1];
    UCHAR       Key3[26+1];
    UCHAR       Key4[26+1];
    UINT        TransRate;
    UINT        TransPower;
    UINT        RTS; 
    UINT        RTSCheck;
    UINT        FragmentCheck;
    UINT        PSmode;
    UINT        PreamType;
    UINT        AntennaRx;
    UINT        AntennaTx;
    UINT        Fragment;
    UCHAR       CountryRegion;
    struct  _RT_PROFILE_SETTING *Next;
} RT_PROFILE_SETTING, *PRT_PROFILE_SETTING;

typedef struct _RT_DEVICE_ADAPTER {
    char    *Device_Name;
    struct _RT_DEVICE_ADAPTER *Next;
} RT_DEVICE_ADAPTER, *PRT_DEVICE_ADAPTER;

//
// IEEE 802.11 Structures and definitions
//
// new types for Media Specific Indications

#define NDIS_802_11_LENGTH_SSID         32
#define NDIS_802_11_LENGTH_RATES        8
#define NDIS_802_11_LENGTH_RATES_EX     16

typedef UCHAR   NDIS_802_11_RATES[NDIS_802_11_LENGTH_RATES];        // Set of 8 data rates

typedef struct _NDIS_802_11_CONFIGURATION_FH
{
   ULONG           Length;            // Length of structure
   ULONG           HopPattern;        // As defined by 802.11, MSB set 
   ULONG           HopSet;            // to one if non-802.11
   ULONG           DwellTime;         // units are Kusec
} NDIS_802_11_CONFIGURATION_FH, *PNDIS_802_11_CONFIGURATION_FH;

typedef struct _NDIS_802_11_CONFIGURATION
{
   ULONG                           Length;             // Length of structure
   ULONG                           BeaconPeriod;       // units are Kusec
   ULONG                           ATIMWindow;         // units are Kusec
   ULONG                           DSConfig;           // Frequency, units are kHz
   NDIS_802_11_CONFIGURATION_FH    FHConfig;
} NDIS_802_11_CONFIGURATION, *PNDIS_802_11_CONFIGURATION;

typedef  ULONG   NDIS_802_11_FRAGMENTATION_THRESHOLD;

typedef  ULONG   NDIS_802_11_RTS_THRESHOLD;

typedef  ULONG   NDIS_802_11_ANTENNA;

typedef enum _RT_802_11_PREAMBLE {
    Rt802_11PreambleLong,
    Rt802_11PreambleShort,
    Rt802_11PreambleAuto
} RT_802_11_PREAMBLE, *PRT_802_11_PREAMBLE;

typedef enum _RT_802_11_PHY_MODE {
    PHY_11BG_MIXED,
    PHY_11B,
    PHY_11G,
    PHY_11A,
    PHY_11ABG_MIXED,
} RT_802_11_PHY_MODE;

typedef struct _RT_802_11_AP_CONFIG {
    ULONG   EnableTxBurst;      // 0-disable, 1-enable
    ULONG   EnableTurboRate;    // 0-disable, 1-enable 72/100mbps turbo rate
    ULONG   IsolateInterStaTraffic;     // 0-disable, 1-enable isolation
    ULONG   HideSsid;           // 0-disable, 1-enable hiding
    ULONG   UseBGProtection;    // 0-AUTO, 1-always ON, 2-always OFF
    ULONG   UseShortSlotTime;   // 0-no use, 1-use 9-us short slot time
    ULONG   Rsv1;               // must be 0
    ULONG   Rsv2;               // must be 0
} RT_802_11_AP_CONFIG, *PRT_802_11_AP_CONFIG;


// structure to define Radius Data
typedef struct _RT_802_11_RADIUS_DATA {
    ULONG RediusIP;     // Radius Server IP Address
    ULONG RediusPort;   // Radius Server Port
    UCHAR RediusKey[64];
}RT_802_11_RADIUS_DATA, *PRT_802_11_RADIUS_DATA;

// structure to define WPA Group Key Rekey Interval
typedef struct _RT_802_11_WPA_REKEY {
    ULONG ReKeyMethod;      // mechanism for rekeying: 0: time-based, 1: packet-based
    ULONG ReKeyInterval;    // time-based: 10seconds, packet-based: kilo-packets
}RT_802_11_WPA_REKEY, *PRT_802_11_WPA_REKEY;

// Add new authentication modes
typedef enum _NDIS_802_11_AUTHENTICATION_MODE
{
    Ndis802_11AuthModeOpen,
    Ndis802_11AuthModeShared,
    Ndis802_11AuthModeAutoSwitch,
    Ndis802_11AuthModeWPA,
    Ndis802_11AuthModeWPAPSK,
    Ndis802_11AuthModeWPANone,
    Ndis802_11AuthModeMax           // Not a real mode, defined as upper bound
} NDIS_802_11_AUTHENTICATION_MODE, *PNDIS_802_11_AUTHENTICATION_MODE;

#define AUTHEN_TYPE_OPEN_SYSTEM     0
#define AUTHEN_TYPE_SHARED_KEY      1
#define AUTHEN_TYPE_802_1X          2
#define AUTHEN_TYPE_WPA             3
#define AUTHEN_TYPE_WPA_PSK         4

// Added new encryption types
// Also aliased typedef to new name
typedef enum _NDIS_802_11_WEP_STATUS
{
    Ndis802_11WEPEnabled,
    Ndis802_11Encryption1Enabled = Ndis802_11WEPEnabled,
    Ndis802_11WEPDisabled,
    Ndis802_11EncryptionDisabled = Ndis802_11WEPDisabled,
    Ndis802_11WEPKeyAbsent,
    Ndis802_11Encryption1KeyAbsent = Ndis802_11WEPKeyAbsent,
    Ndis802_11WEPNotSupported,
    Ndis802_11EncryptionNotSupported = Ndis802_11WEPNotSupported,
    Ndis802_11Encryption2Enabled,
    Ndis802_11Encryption2KeyAbsent,
    Ndis802_11Encryption3Enabled,
    Ndis802_11Encryption3KeyAbsent
} NDIS_802_11_WEP_STATUS, *PNDIS_802_11_WEP_STATUS,
  NDIS_802_11_ENCRYPTION_STATUS, *PNDIS_802_11_ENCRYPTION_STATUS;

typedef  ULONG  NDIS_802_11_KEY_INDEX;
typedef ULONGLONG   NDIS_802_11_KEY_RSC;

// Key mapping keys require a BSSID
typedef struct _NDIS_802_11_KEY
{
    ULONG           Length;     // Length of this structure
    ULONG           KeyIndex;
    ULONG           KeyLength;  // length of key in bytes
    NDIS_802_11_MAC_ADDRESS BSSID;
    NDIS_802_11_KEY_RSC KeyRSC;
    UCHAR           KeyMaterial[1];     // variable length depending on above field
} NDIS_802_11_KEY, *PNDIS_802_11_KEY;

typedef struct _NDIS_802_11_REMOVE_KEY
{
    ULONG                   Length;        // Length of this structure
    ULONG                   KeyIndex;
    NDIS_802_11_MAC_ADDRESS BSSID;
} NDIS_802_11_REMOVE_KEY, *PNDIS_802_11_REMOVE_KEY;


#define MAX_LEN_OF_MAC_TABLE        32
#define ETH_LENGTH_OF_ADDRESS       6


typedef struct _NDIS_802_11_SSID 
{
    ULONG   SsidLength;         // length of SSID field below, in bytes;
                                // this can be zero.
    UCHAR   Ssid[NDIS_802_11_LENGTH_SSID];      // SSID information field
} NDIS_802_11_SSID, *PNDIS_802_11_SSID;

typedef struct _RT_802_11_ACL_ENTRY {
    UCHAR   Addr[ETH_LENGTH_OF_ADDRESS];
    USHORT  Rsv;
} RT_802_11_ACL_ENTRY, *PRT_802_11_ACL_ENTRY;

typedef struct _RT_802_11_ACL {
    ULONG   Policy;             // 0-disable, 1-positive list, 2-negative list
    ULONG   Num;
    RT_802_11_ACL_ENTRY Entry[MAX_LEN_OF_MAC_TABLE];
} RT_802_11_ACL, *PRT_802_11_ACL;

// put all proprietery for-query objects here to reduce # of Query_OID
typedef struct _RT_802_11_LINK_STATUS {
    ULONG   CurrTxRate;         // in units of 0.5Mbps
    ULONG   ChannelQuality;     // 0..100 %
    ULONG   TxByteCount;        // both ok and fail
    ULONG   RxByteCount;        // both ok and fail
} RT_802_11_LINK_STATUS, *PRT_802_11_LINK_STATUS;

//
// Defines the state of the LAN media
//
typedef enum _NDIS_MEDIA_STATE
{
    NdisMediaStateConnected,
    NdisMediaStateDisconnected
} NDIS_MEDIA_STATE, *PNDIS_MEDIA_STATE;

//
// Received Signal Strength Indication
//
typedef LONG    NDIS_802_11_RSSI;           // in dBm

typedef struct _RT_802_11_MAC_ENTRY {
    UCHAR       Addr[ETH_LENGTH_OF_ADDRESS];
    UCHAR       Aid;
    UCHAR       Psm;     // 0:PWR_ACTIVE, 1:PWR_SAVE
} RT_802_11_MAC_ENTRY, *PRT_802_11_MAC_ENTRY;

typedef struct _RT_802_11_MAC_TABLE {
    ULONG       Num;
    RT_802_11_MAC_ENTRY Entry[MAX_LEN_OF_MAC_TABLE];
} RT_802_11_MAC_TABLE, *PRT_802_11_MAC_TABLE;

#define MAX_NUM_OF_EVENT               10  // entry # in EVENT table
#define MAX_ARRAY_SIZE_OF_EVENT        100 // entry # in EVENT table

typedef struct _RT_802_11_EVENT_LOG {
    ULONG           SystemTime;  // timestammp via NdisGetCurrentSystemTime()
    UCHAR           Addr[ETH_LENGTH_OF_ADDRESS];
    USHORT          Event;       // EVENT_xxx
} RT_802_11_EVENT_LOG, *PRT_802_11_EVENT_LOG;

typedef struct _RT_802_11_EVENT_TABLE {
    ULONG       Num;
    ULONG       Rsv;    // to align Log[] at LARGE_INEGER boundary
    RT_802_11_EVENT_LOG   Log[MAX_NUM_OF_EVENT];
} RT_802_11_EVENT_TABLE, PRT_802_11_EVENT_TABLE;

typedef struct _NDIS_802_11_STATISTICS
{
   ULONG           Length;      // Length of structure
   LARGE_INTEGER   TransmittedFragmentCount;
   LARGE_INTEGER   MulticastTransmittedFrameCount;
   LARGE_INTEGER   FailedCount;
   LARGE_INTEGER   RetryCount;
   LARGE_INTEGER   MultipleRetryCount;
   LARGE_INTEGER   RTSSuccessCount;
   LARGE_INTEGER   RTSFailureCount;
   LARGE_INTEGER   ACKFailureCount;
   LARGE_INTEGER   FrameDuplicateCount;
   LARGE_INTEGER   ReceivedFragmentCount;
   LARGE_INTEGER   MulticastReceivedFrameCount;
   LARGE_INTEGER   FCSErrorCount;
} NDIS_802_11_STATISTICS, *PNDIS_802_11_STATISTICS;

typedef struct _RT_VERSION_INFO{
    UCHAR       DriverMajorVersion;
    UCHAR       DriverMinorVersion;
    UCHAR       DriverSubVersion;
    UINT        DriverBuildYear;
    UINT        DriverBuildMonth;
    UINT        DriverBuildDay;
} RT_VERSION_INFO, *PRT_VERSION_INFO;


typedef struct _RT_802_11_WDS {
    ULONG   Num;
    NDIS_802_11_MAC_ADDRESS MacAddress[3];
    ULONG    KeyLength;
    UCHAR    KeyMaterial[32];
} RT_802_11_WDS, *PRT_802_11_WDS;
////////////////////////////////////////////////////////////////////////////////////////////
void AtoH(char * src, UCHAR * dest, int destlen);
static UCHAR BtoH(char ch);
int Open_Socket(void);
int OidQueryInformation(USHORT OidQueryCode, int socket_id, char *DeviceName, void *ptr, ULONG PtrLength);
int OidSetInformation(USHORT OidQueryCode, int socket_id, char *DeviceName, void *ptr, ULONG PtrLength);
int PasswordHash(char *password, unsigned char *ssid, int ssidlength, unsigned char *output);

unsigned int  ConvertRssiToSignalQuality(int RSSI);

void unmktime(unsigned long time, long offset,
            int *yearp, int *monp, int *dayp,
            int *hourp, int *minp, int *secp);

#endif //RT_TOOL_H

